"""Top-level package for sarwaveifrproc."""

__author__ = """Antoine Grouazel"""
__email__ = "antoine.grouazel@ifremer.fr"

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata
__version__ = metadata.version("sarwaveifrproc")



import os

import pytest

from sarwaveifrproc.utils import SAFE_PATTERN, get_output_filename, get_output_safe

inputs_l1slc = [
    "/tmp/2022/127/S1A_IW_XSP__1SDV_20220507T162437_20220507T162504_043107_0525DE_B14E_A02.SAFE/l1b-s1a-iw1-xsp-vv-20220507t162439-20220507t162504-043107-0525de-004-a02.nc",
    "/tmp/2022/127/S1A_IW_XSP__1SDV_20220507T162437_20220507T162504_043107_0525DE_B14E_A02.SAFE/l1b-s1a-iw1-xsp-vv-20220507t162439-20220507t162504-043107-0525de-004_a02.nc",
    "/tmp/2022/127/S1A_IW_XSP__1SDV_20220507T162437_20220507T162504_043107_0525DE_B14E_B02.SAFE/l1c-s1a-iw1-xsp-vv-20220507t162439-20220507t162504-043107-0525de-004_b02.nc",
    "/tmp/2022/127/S1A_IW_XSP__1SDV_20220507T162437_20220507T162504_043107_0525DE_B14E.SAFE/l1c-s1a-iw1-xsp-vv-20220507t162439-20220507t162504-043107-0525de-004_b02.nc",
]
expected_l2wav = [
    "/tmp/2022/127/S1A_IW_WAV__2SDV_20220507T162437_20220507T162504_043107_0525DE_B14E_E03.SAFE/l2-s1a-iw1-wav-dv-20220507t162439-20220507t162504-043107-0525de-e03.nc"
]


@pytest.mark.parametrize(
    ["l1b_fullpath", "expected_l2wav"],
    (
        pytest.param(inputs_l1slc[0], expected_l2wav[0]),
        pytest.param(inputs_l1slc[1], expected_l2wav[0]),
        pytest.param(inputs_l1slc[2], expected_l2wav[0]),
        pytest.param(inputs_l1slc[3], expected_l2wav[0]),
    ),
)
def test_outputfile_path(l1b_fullpath, expected_l2wav):
    version = "E03"
    outputdir = "/tmp/"
    output_safe = get_output_safe(
        l1x_safe=os.path.dirname(l1b_fullpath), root_savepath=outputdir, tail=version
    )
    l2_full_path = get_output_filename(l1b_fullpath, output_safe, tail=version)

    print(l2_full_path)
    assert l2_full_path == expected_l2wav


if __name__ == "__main__":
    import re

    for path in inputs_l1slc:
        safe_name = os.path.basename(os.path.dirname(path))
        match = re.match(SAFE_PATTERN, safe_name)
        if match:
            print("✔ Match:", safe_name, match.groupdict())
        else:
            print("❌ No match:", safe_name)



import glob
import logging
import os
import re
import traceback
from datetime import datetime

import xarray as xr

from sarwaveifrproc.l2_wave import generate_l2_wave_product

SAFE_PATTERN = (
    r"^(?P<mission_id>S1[A-Z])_"
    r"(?P<mode>[IE]W)_"
    r"(?P<type>[A-Z]{3})__"
    r"(?P<level>[0-9])(?P<class>[A-Z])(?P<pol>[A-Z]{2})_"
    r"(?P<starttime>\d{8}T\d{6})_"
    r"(?P<endtime>\d{8}T\d{6})_"
    r"(?P<orbit_no>\d{6})_"
    r"(?P<datatake_id>[A-Z0-9]{6})_"
    r"(?P<id>[A-Z0-9]{4})"
    r"(?:_(?P<suffix>[A-Z0-9]{3}))?"  # optional _B02, _A02, etc.
    r"\.SAFE$"
)


def get_safe_date(safe):
    """
    Extracts and returns the date and time from a given SAFE directory name.

    Parameters:
        safe (str): SAFE directory name.

    Returns:
        datetime: A datetime object representing the extracted date and time.
    """
    date_str = safe.split("_")[5]
    date = datetime.strptime(date_str, "%Y%m%dT%H%M%S")
    return date


def get_output_safe(l1x_safe, root_savepath, tail="E00"):
    """
    Generates the final SAFE filename with specified modifications.

    Parameters:
        l1x_safe (str): The input SAFE path.
        root_savepath (str): The root path for saving the output.
        tail (str): The tail to append to the filename. Defaults to 'E00'.

    Returns:
        str: The output savepath.

    Raises:
        ValueError: If the input SAFE does not meet the filename requirements.
    """
    l1x_safe = l1x_safe.rstrip("/")  # remove trailing slash
    safe = l1x_safe.split(os.sep)[-1]
    final_safe = safe
    info = re.match(SAFE_PATTERN, final_safe)
    m = info.groupdict()
    cond1 = m["type"] == "XSP"
    cond2 = m["pol"] in ["SV", "DV"]
    cond3 = final_safe.endswith(".SAFE")
    if cond1 and cond2 and cond3:
        date = get_safe_date(safe)

        final_safe = final_safe.replace("XSP_", "WAV_")
        final_safe = final_safe.replace("1SDV", "2SDV")
        final_safe = final_safe.replace("1SSV", "2SSV")
        regexA = re.compile("A[0-9]{2}.SAFE")
        regexB = re.compile("B[0-9]{2}.SAFE")
        if re.search(regexA, final_safe.split("_")[-1]) or re.search(
            regexB, final_safe.split("_")[-1]
        ):
            final_safe = final_safe.replace(
                final_safe.split("_")[-1], f"{tail.upper()}.SAFE"
            )
        else:
            print("no slug existing-> just add the product ID")
            final_safe = final_safe.replace(".SAFE", f"_{tail.upper()}.SAFE")

        output_safe = os.path.join(
            root_savepath, date.strftime("%Y"), date.strftime("%j"), final_safe
        )

        return output_safe

    else:
        raise ValueError("The input SAFE does not meet the filename requirements.")


def get_output_filename(l1x_path, output_safe, tail="e00"):
    """
    Constructs and returns the file path for saving a processed file based on input parameters.

    Parameters:
        l1x_path (str): The path to the input file. It can be either a L1B or L1C file.
        root_savepath (str): The root directory where the processed file will be saved.
        tail (str, optional): The tail string to be appended to the filename corresponding to the processing options. Defaults to 'e00'.

    Returns:
        str: File path for saving the processed file.
    """
    filename = l1x_path.split(os.sep)[-1]
    filename_exploded = filename.split("-")
    regex_file_number = re.compile("[0-9]{2}[0-6]")
    if filename[0:2] == "l1":
        final_filename = "-".join(
            [
                *filename_exploded[:4],
                "dv",
                *filename_exploded[5:-1],
                f"{tail.lower()}.nc",
            ]
        )
        if re.search(regex_file_number, final_filename.split("-")[9]):
            final_filename = final_filename.replace(
                final_filename.split("-")[9] + "-", ""
            )  # remove the -004- giving the number of the file
    else:
        final_filename = "-".join(
            [
                *filename_exploded[:3],
                "dv",
                *filename_exploded[4:-1],
                f"{tail.lower()}.nc",
            ]
        )
        if re.search(regex_file_number, final_filename.split("-")[8]):
            final_filename = final_filename.replace(
                final_filename.split("-")[8] + "-", ""
            )  # remove the -004- giving the number of the file
    final_filename = final_filename.replace("l1b", "l2")
    final_filename = final_filename.replace("l1c", "l2")
    final_filename = final_filename.replace("xsp", "wav")
    savepath = os.path.join(output_safe, final_filename)
    return savepath


def process_files(
    input_safe, output_safe, models, models_outputs, predicted_variables, product_id
):
    """
    Processes files in the input directory, generates predictions, and saves results in the output directory.

    Parameters:
        input_safe (str): Input safe path.
        output_safe (str): Path to the directory where output data will be saved.
        models (dict): dict of onnx runtime inference sessions
        models_outputs (dict): dict of List of model outputs names
        predicted_variables (list): List of variable names to be predicted.
        product_id (str): Identifier for the output product.
    Returns:
        files_in_error (list): List of files that encountered errors during processing.
    """
    subswath_filenames = glob.glob(os.path.join(input_safe, "*?v*.nc"))
    logging.debug(f"{len(subswath_filenames)} subswaths found in given safe.")
    files_in_error = []
    for path in subswath_filenames:
        try:
            xdt = xr.DataTree.from_dict(xr.open_groups(path))
            l2_product = generate_l2_wave_product(
                xdt, models, models_outputs, predicted_variables
            )

            os.makedirs(output_safe, exist_ok=True)
            savepath = get_output_filename(path, output_safe, product_id)
            l2_product.to_netcdf(savepath)
        except Exception:
            logging.errror(traceback.format_exc())
            logging.error(f"Error processing {path}. Skipping this file.")
            files_in_error.append(path)
            continue
    return files_in_error  # can be used to reprocess the files in failure



from pathlib import Path

import hydra_zen

import sarwaveifrproc.main


def chunk_listing(path: str, i: int, n: int = 10) -> str:
    ll = Path(path).read_text().split()
    chunk_size = len(ll) // (n - 1)
    Path("tmp").mkdir(exist_ok=True)
    Path(f"tmp/chunk_{i}_{Path(path).name}").write_text(
        "\n".join(ll[i * chunk_size : (i + 1) * chunk_size])
    )
    return f"tmp/chunk_{i}_{Path(path).name}"


hydra_zen.store(sarwaveifrproc.main.main, name="base")
hydra_zen.store(
    dict(header=sarwaveifrproc.main.main.__doc__),
    name="doc",
    group="hydra/help",
)

hydra_zen.store(
    dict(
        input_path=hydra_zen.builds(
            chunk_listing,
            path="???",
            i="${hydra:job.num}",
            n="${hydra:launcher.n_jobs}",
        )
    ),
    name="chunk",
    group="parallel",
    package="_global_",
)

hydra_zen.store.add_to_hydra_store(overwrite_ok=True)



import logging
import os
import re
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import hydra
import hydra_zen
import numpy as np
import onnxruntime
from tqdm import tqdm

import sarwaveifrproc.utils as utils


@dataclass
class Model:
    """
    path: path to the onnx file
    outputs: names of the model output variables
    """

    path: str
    outputs: list[str]


@dataclass
class Prediction:
    """
    model: name of the model that predict this variable
    output_name: name of the model output
    attrs: attribute dictionnary of the variable
    """

    model: str
    output: str
    attrs: dict[str, str]


@dataclass
class PredictedVariables:
    """
    intraburst: variables to add in the intraburst
    interburst: variables to add in the interburst
    """

    intraburst: dict[str, Prediction]
    interburst: dict[str, Prediction]


def main(
    input_path,
    save_directory: str,
    product_id: str,
    models: dict[str, Model],
    predicted_variables: PredictedVariables,
    supported_input_product_versions: list[str] = [],
    overwrite: bool = False,
    verbose: bool = False,
    dry_run: bool = False,
):
    """
    Generate a L2 WAVE product from a L1B or L1C SAFE.

    input_path: l1b or l1c safe path or listing path (.txt file).
    save_directory: where to save output data
    product_id: 3 digits ID representing the processing options. Ex: E00.
    models: onnx models and output
    predicted_variables:  model outputs and associated variables name to add to the L2 product
    overwrite: overwrite the existing outputs
    verbose: debug log level if True
    supported_input_product_versions: list of product versions the model exlicitely supports
    dry_run: flag to skip the actual processing
    """

    setup_logging(verbose)
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = (
        "3"  # Suppress TensorFlow INFO and WARNING messages
    )
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # Hide CUDA devices
    logging.info("Loading configuration file...")

    logging.info("Loading models...")
    ort_mods = {k: onnxruntime.InferenceSession(d.path) for k, d in models.items()}
    mod_outs = {k: d.outputs for k, d in models.items()}
    logging.info("Models loaded.")

    if input_path.endswith(".txt"):
        files = np.loadtxt(input_path, dtype=str)
        output_safes = np.array(
            [utils.get_output_safe(f, save_directory, product_id) for f in files]
        )

        if not overwrite:
            mask = np.array([os.path.exists(f) for f in output_safes])
            files, output_safes = files[~mask], output_safes[~mask]

            logging.info(
                f"{np.sum(mask)} file(s) already exist(s) and overwriting is not allowed. Use --overwrite to overwrite existing files."
            )

            if not files.size:
                return None

        logging.info("Processing files...")
        for iix in tqdm(range(len(files))):
            f = files[iix]
            output_safe = output_safes[iix]
            name = Path(f).name
            m = re.match(utils.SAFE_PATTERN, name)
            if (
                m is None
                or m.groupdict().get("version") not in supported_input_product_versions
            ) and iix == 0:
                logging.warning(f"Unsupported product version for SAFE {name}")
            if dry_run:
                continue
            files_in_error = utils.process_files(
                f, output_safe, ort_mods, mod_outs, predicted_variables, product_id
            )
            if len(files_in_error) > 0:
                filout_error = os.path.join(
                    "files_in_error_{}.txt".format(
                        datetime.now().strftime("%Y%m%d_%H%M%S")
                    )
                )
                logging.info(f"writting the list of files in error to  {filout_error}")
                with open(filout_error, "w") as ff:
                    ff.writelines("\n".join(files_in_error))
            else:
                logging.info("All provided SAFE have been processed successfuly.")

    else:
        name = Path(input_path).name
        m = re.match(utils.SAFE_PATTERN, name)
        if (
            m is None
            or m.groupdict().get("version") not in supported_input_product_versions
        ):
            logging.warning(f"Unsupported product version for SAFE {name}")
        logging.info("Checking if output safe already exists...")
        output_safe = utils.get_output_safe(input_path, save_directory, product_id)

        if os.path.exists(output_safe) and not overwrite:
            logging.info(
                f"{output_safe} already exists and overwriting is not allowed. Use --overwrite to overwrite existing files."
            )
            return None

        logging.info("Processing files...")
        if not dry_run:
            files_in_error = utils.process_files(
                input_path,
                output_safe,
                ort_mods,
                mod_outs,
                predicted_variables,
                product_id,
            )

    logging.info(f"Processing terminated. Output directory: \n{save_directory}")


def setup_logging(verbose=False):
    fmt = "%(asctime)s %(levelname)s %(filename)s(%(lineno)d) %(message)s"
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=level, format=fmt, datefmt="%d/%m/%Y %H:%M:%S", force=True
    )


hydra_main = hydra.main(
    config_name="e11",
    config_path="pkg://sarwave_config",
    version_base="1.3",
)(hydra_zen.zen(main))



import logging
import os

import numpy as np
import onnxruntime
import pytest
import xarray as xr
from hydra import compose, initialize

import sarwaveifrproc
from sarwaveifrproc.l2_wave import generate_intermediate_product

VARIABLE = "hs_most_likely"
GRP = "intraburst"
kept_variables = [
    "corner_longitude",
    "corner_latitude",
    "land_flag",
    "sigma0_filt",
    "normalized_variance_filt",
    "incidence",
    "azimuth_cutoff",
    "cwave_params",
]


# reference prediction is E01 (rmarquart January 2024), validated by L.Maillard
def get_hs_values(ff):
    ds = xr.open_dataset(ff, group=GRP)
    print("hs?", [uu for uu in ds if "hs" in uu])
    return ds[VARIABLE].data


# expected outputs
subswathSAR = [
    os.path.join(
        os.path.dirname(sarwaveifrproc.__file__),
        "reference_data",
        # "s1a-iw2-slc-dv-20231128t035702-20231128t035727-051412-063451-e01.nc",
        # "l2-s1a-iw2-wav-dv-20220307t015815-20220307t015843-042209-0507b6-e11.nc"
        "l2-s1b-iw2-wav-dv-20210311t055224-20210311t055250-025963-0318d0-e11.nc",
    ),
]
# L1B_SAR_vh = [os.path.join(os.path.dirname(sarwaveifrproc.__file__),'reference_data','s1a-iw2-slc-vh-20231128t035702-20231128t035727-051412-063451-002_L1B_xspec_IFR_3.7.6nospectra.nc')]
L1B_SAR_vv = [
    os.path.join(
        os.path.dirname(sarwaveifrproc.__file__),
        "reference_data",
        # "s1a-iw2-slc-vv-20231128t035702-20231128t035727-051412-063451-005_L1B_xspec_IFR_3.7_nospectra.nc",
        "l1c-s1b-iw2-xsp-1sdv-20210311t055224-20210311t055250-025963-0318d0-b09.nc",
    )
]
hs_expected = []
for ii in subswathSAR:
    # hs_expected[ii] = get_hs_values(ff=ii)
    hs_expected.append(get_hs_values(ff=ii))


@pytest.mark.parametrize(
    "L1B_SAR_vv, hs_expected",
    [(L1B_SAR_vv[0], hs_expected[0])],
)
def test_hs_prediction_E11(L1B_SAR_vv, hs_expected):
    # Manually initialize Hydra
    initialize(config_path="pkg://sarwave_config", version_base="1.3")
    cfg = compose(config_name="e11")
    models = cfg["models"]
    xdt = xr.DataTree.from_dict(xr.open_groups(L1B_SAR_vv))
    logging.info("Loading models...")
    ort_mods = {k: onnxruntime.InferenceSession(d.path) for k, d in models.items()}
    mod_outs = {k: d.outputs for k, d in models.items()}
    logging.info("Models loaded.")
    predicted_variables = cfg["predicted_variables"]
    ds_intraburst = generate_intermediate_product(
        xdt[GRP].ds,
        models=ort_mods,
        models_outputs=mod_outs,
        predicted_variables=predicted_variables.intraburst,
        kept_variables=kept_variables,
        pol="VV",
    )
    print("ds_intraburst", ds_intraburst)

    actual_hs_values = ds_intraburst[VARIABLE].data
    print(actual_hs_values.shape)
    print(hs_expected.shape)
    print(actual_hs_values, hs_expected)
    assert np.allclose(actual_hs_values, hs_expected, atol=1e-03, equal_nan=True)


if __name__ == "__main__":
    import argparse
    import logging

    parser = argparse.ArgumentParser(description="testHsinference")
    parser.add_argument("--verbose", action="store_true", default=False)
    parser.add_argument(
        "--input",
        required=False,
        default=L1B_SAR_vv[0],
        help="path of the l1b measurement to test (s1a-iw2-slc-vv-20231128t035702-20231128t035727-051412-063451-005)",
    )
    args = parser.parse_args()
    fmt = "%(asctime)s %(levelname)s %(filename)s(%(lineno)d) %(message)s"
    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG, format=fmt, datefmt="%d/%m/%Y %H:%M:%S", force=True
        )
    else:
        logging.basicConfig(
            level=logging.INFO, format=fmt, datefmt="%d/%m/%Y %H:%M:%S", force=True
        )
    logging.info("start")
    test_hs_prediction_E11(args.input, hs_expected[0])



import os

import numpy as np
import xarray as xr

import sarwaveifrproc

attributes_missing_variables = {
    "sigma0_filt": {
        "long_name": "calibrated sigma0 with BT correction",
        "units": "linear",
    },
    "normalized_variance_filt": {
        "long_name": "normalized variance with BT correction",
        "units": "",
    },
    "azimuth_cutoff": {"long_name": "Azimuthal cut-off (2tau)", "units": "m"},
    "cwave_params": {"long_name": "CWAVE parameters"},
    "k_gp": {"long_name": "Gegenbauer polynoms dimension"},
    "phi_hf": {"long_name": "Harmonic functions dimension (odd number)"},
}


def generate_l2_wave_product(xdt, models, models_outputs, predicted_variables):
    """
    Generate a level-2 wave (L2 WAV) product.

    Parameters:
    - xdt (dict): DataTree containing intraburst and interburst datasets.
    - models (dict[str, onnxruntime.InferenceSession): different ml models used
    - models_outputs (dict[str, list]): list of variables predicted by each model
    - predicted_variables (dict[dict]):  variables to add to the product and corresponding model and output name
    Returns:
    - l2_wave_product (xr.DataTree): Level-2 wave product.

    Notes:
    - The scaler objects should be one of StandardScaler, MinMaxScaler, or RobustScaler from sklearn.preprocessing.
    """

    kept_variables = [
        "corner_longitude",
        "corner_latitude",
        "land_flag",
        "sigma0_filt",
        "normalized_variance_filt",
        "incidence",
        "azimuth_cutoff",
        "cwave_params",
    ]
    intraburst_models = {
        k: m
        for k, m in models.items()
        if k in [v.model for v in predicted_variables.intraburst.values()]
    }
    interburst_models = {
        k: m
        for k, m in models.items()
        if k in [v.model for v in predicted_variables.interburst.values()]
    }
    ds_intraburst = generate_intermediate_product(
        xdt["intraburst"].ds,
        intraburst_models,
        models_outputs,
        predicted_variables.intraburst,
        kept_variables,
    )
    ds_interburst = generate_intermediate_product(
        xdt["interburst"].ds,
        interburst_models,
        models_outputs,
        predicted_variables.interburst,
        kept_variables,
    )

    l2_wave_product = xr.DataTree.from_dict(
        {"intraburst": ds_intraburst, "interburst": ds_interburst}
    )
    l2_wave_product.attrs["input_sar_product"] = os.path.basename(
        xdt.encoding["source"]
    )
    return l2_wave_product


def generate_intermediate_product(
    ds, models, models_outputs, predicted_variables, kept_variables, pol="VV"
):
    """
    Generate an intermediate l2 product, depending of the input dataset (intraburst or interburst).

    Parameters:
    - ds (xarray.Dataset): Input dataset.
    - models (dict[str, onnxruntime.InferenceSession): different ml models used
    - models_outputs (dict[str, list]): list of variables predicted by each model
    - predicted_variables (dict[dict]):  variables to add to the product and corresponding model and output name
    - kept_variables (list): List of variables from the input dataset that are kept in the final product.
    - pol (str): polarisation to select

    Returns:
    - ds_pred (xarray.Dataset): Intermediate predictions dataset.
    """

    if ds["land_flag"].all():
        return generate_product_on_land(ds, predicted_variables, kept_variables)

    if "2tau" in ds.dims:
        ds = ds.squeeze(dim="2tau")
        ds.attrs["squeezed_dimensions"] = "2tau"

    tiles = ds[
        [
            "sigma0_filt",
            "normalized_variance_filt",
            "incidence",
            "azimuth_cutoff",
            "cwave_params",
        ]
    ].sel(pol=pol)
    if "burst" in ds.coords:
        tiles_stacked = tiles.stack(
            all_tiles=["burst", "tile_line", "tile_sample"], k_phi=["k_gp", "phi_hf"]
        )
    else:
        tiles_stacked = tiles.stack(
            all_tiles=["tile_line", "tile_sample"], k_phi=["k_gp", "phi_hf"]
        )

    output_dims = [["preds", "all_tiles"]]

    predictions = xr.concat(
        [
            xr.apply_ufunc(
                predict_variables,
                tiles_stacked["sigma0_filt"],
                tiles_stacked["normalized_variance_filt"],
                tiles_stacked["incidence"],
                tiles_stacked["azimuth_cutoff"],
                tiles_stacked["cwave_params"],
                model,
                input_core_dims=[
                    ["all_tiles"],
                    ["all_tiles"],
                    ["all_tiles"],
                    ["all_tiles"],
                    ["all_tiles", "k_phi"],
                    [],
                ],
                output_core_dims=output_dims,
                vectorize=False,
            ).assign_coords(preds=[f"{k}_{o}" for o in models_outputs[k]])
            for (k, model) in models.items()
        ],
        dim="preds",
    )

    ds_pred = format_dataset(ds, predictions, predicted_variables, kept_variables)

    return ds_pred


def generate_product_on_land(ds, predicted_variables, kept_variables):
    """
    Patch function when input dataset does not contain all necessary variables.

    Parameters:
    - ds (xarray.Dataset): Input dataset.
    - predicted_variables (list): List of predicted variable names.
    - kept_variables (list): List of variables from the input dataset that are kept in the final product.
    - bins (dict): Dictionary containing bins for each variable.
    """
    ds = ds[["corner_longitude", "corner_latitude", "incidence", "land_flag"]]

    shape = ds["land_flag"].shape
    dims = ds["land_flag"].dims

    data_to_merge = []

    if set(predicted_variables).issubset(ds.keys()):
        kept_variables = kept_variables + predicted_variables

    for v in kept_variables:
        if v not in ds.variables:
            v_array = xr.DataArray(data=np.full(shape, np.nan), dims=dims).rename(v)
            v_array.attrs = attributes_missing_variables[v]
            data_to_merge.append(v_array)

    # Add CWAVES independently because the required dimensions are not in the input dataset when there is only land.
    k_gp = xr.DataArray(data=range(1, 5), dims="k_gp")
    k_gp.attrs = attributes_missing_variables["k_gp"]
    phi_hf = xr.DataArray(data=range(1, 6), dims="phi_hf")
    phi_hf.attrs = attributes_missing_variables["phi_hf"]

    cwaves = xr.DataArray(
        data=np.full(shape + (4, 5), np.nan), dims=dims + ("k_gp", "phi_hf")
    ).rename("cwave_params")
    cwaves.attrs = attributes_missing_variables["cwave_params"]
    data_to_merge.append(cwaves.assign_coords({"k_gp": k_gp, "phi_hf": phi_hf}))

    for v, vd in predicted_variables.items():
        attributes = vd.attrs
        preds = xr.DataArray(data=np.full(shape, np.nan), dims=dims).rename(v)
        preds.attrs = attributes
        data_to_merge.append(preds)

    ds = xr.merge([ds] + data_to_merge)

    ds.attrs.pop("name", None)
    ds.attrs.pop("multidataset", None)
    return ds


def predict_variables(
    sigma0, normalized_variance, incidence, azimuth_cutoff, cwave_params, model
):
    """
    Launch predictions using a neural model.

    Parameters:
    - sigma0 (xarray.DataArray): Array containing sigma0 values.
    - normalized_variance (xarray.DataArray): Array containing normalized variance values.
    - incidence (xarray.DataArray): Array containing incidence values.
    - azimuth_cutoff (xarray.DataArray): Array containing azimuth cutoff values.
    - cwave_params (xarray.DataArray): Array containing the cwave parameters.
    - model (tf.keras.Model): Machine learning model for predicting sea parameters.

    Returns:
    - res (tuple): Tuple containing predictions for each variable.
    """
    X_stacked = np.vstack([sigma0, normalized_variance, incidence, azimuth_cutoff]).T
    X_stacked = np.hstack([cwave_params, X_stacked])
    # X_normalized = scaler.transform(X_stacked)

    input_name = model.get_inputs()[0].name
    inputs = {input_name: X_stacked.astype(np.float32)}

    res = model.run(None, inputs)
    res = [r[:, i] for r in res for i in range(r.shape[-1])]
    return res


def format_dataset(ds, predictions, predicted_variables, kept_variables):
    """
    Format a dataset based on predictions, variables, and bins.

    Parameters:
    - ds (xarray.Dataset): Input dataset (l1b or l1c).
    - predictions (list): Predictions for each variable.
    - predicted_variables (list): List of predicted variable names.
    - kept_variables (list): List of variables from the input dataset that are kept in the final product.
    - bins (dict): Dictionary containing bin values for each variable.

    Returns:
    - ds_pred (xarray.Dataset): Formatted dataset with predictions and original data.

    The function computes various statistical values (mean, most likely, standard deviation)
    from the predictions and adds them to the new dataset. It also includes original data
    from the input dataset and some additional attributes.
    """

    data_to_merge = []

    for v, vd in predicted_variables.items():
        attributes = vd.attrs
        preds = predictions.sel(preds=f"{vd.model}_{vd.output}", drop=True).rename(v)
        preds.attrs = attributes
        data_to_merge.append(preds)

    if set(predicted_variables).issubset(ds.keys()):
        kept_variables = kept_variables + predicted_variables

    data_to_merge = [ds.unstack() for ds in data_to_merge]
    ds = ds[kept_variables]

    ds = xr.merge([ds] + data_to_merge)
    ds = ds.drop_vars(["tile_line", "tile_sample"])
    ds.attrs["l2_processor_name"] = "sarwaveifrproc"
    ds.attrs["l2_processor_version"] = sarwaveifrproc.__version__
    ds.attrs.pop("name", None)
    ds.attrs.pop("multidataset", None)

    return ds
